from __future__ import annotations

import dataclasses
import datetime
import ipaddress
import json
import time
import typing

import quart

import capport.utils.zoneinfo

if typing.TYPE_CHECKING:
    from .config import Config


IPAddress = ipaddress.IPv4Address | ipaddress.IPv6Address


@dataclasses.dataclass(frozen=True)
class MacAddress:
    raw: bytes

    def __str__(self) -> str:
        return self.raw.hex(":")

    def __repr__(self) -> str:
        return repr(str(self))

    @staticmethod
    def parse(s: str) -> MacAddress:
        return MacAddress(bytes.fromhex(s.replace(":", "")))


@dataclasses.dataclass(frozen=True, order=True)
class Timestamp:
    epoch: int

    def __str__(self) -> str:
        try:
            ts = datetime.datetime.fromtimestamp(self.epoch, datetime.timezone.utc)
            return ts.isoformat(sep=" ")
        except OSError:
            return f"epoch@{self.epoch}"

    def __repr__(self) -> str:
        return repr(str(self))

    @staticmethod
    def now() -> Timestamp:
        now = int(time.time())
        return Timestamp(epoch=now)

    @staticmethod
    def from_protobuf(epoch: int) -> Timestamp | None:
        if epoch:
            return Timestamp(epoch=epoch)
        return None


@dataclasses.dataclass
class MacPublicState:
    address: IPAddress
    mac: MacAddress | None
    allowed_remaining: int

    @staticmethod
    def from_missing_mac(address: IPAddress) -> MacPublicState:
        return MacPublicState(
            address=address,
            mac=None,
            allowed_remaining=0,
        )

    @property
    def allowed(self) -> bool:
        return self.allowed_remaining > 0

    @property
    def captive(self) -> bool:
        return not self.allowed

    @property
    def allowed_remaining_duration(self) -> str:
        mm, ss = divmod(self.allowed_remaining, 60)
        hh, mm = divmod(mm, 60)
        return f"{hh}:{mm:02}:{ss:02}"

    @property
    def allowed_until(self) -> datetime.datetime | None:
        zone = capport.utils.zoneinfo.get_local_timezone()
        now = datetime.datetime.now(zone).replace(microsecond=0)
        return now + datetime.timedelta(seconds=self.allowed_remaining)

    def to_rfc8908(self, config: Config) -> quart.Response:
        response: dict[str, typing.Any] = {
            "user-portal-url": quart.url_for("index", _external=True),
        }
        if config.venue_info_url:
            response["venue-info-url"] = config.venue_info_url
        if self.captive:
            response["captive"] = True
        else:
            response["captive"] = False
            response["seconds-remaining"] = self.allowed_remaining
            response["can-extend-session"] = True
        return quart.Response(
            json.dumps(response),
            headers={"Cache-Control": "private"},
            content_type="application/captive+json",
        )
