from __future__ import annotations

import contextlib
import errno
import ipaddress
import socket
import typing

import pyroute2.iproute.linux  # type: ignore
import pyroute2.netlink.exceptions  # type: ignore
import pyroute2.netlink.rtnl  # type: ignore
import pyroute2.netlink.rtnl.ndmsg  # type: ignore

from capport import cptypes


@contextlib.asynccontextmanager
async def connect():
    yield NeighborController()


# TODO: run blocking iproute calls in a different thread?
class NeighborController:
    def __init__(self):
        self.ip = pyroute2.iproute.linux.IPRoute()

    async def get_neighbor(
        self,
        address: cptypes.IPAddress,
        *,
        index: int = 0,  # interface index
        flags: int = 0,
    ) -> pyroute2.iproute.linux.ndmsg.ndmsg | None:
        if not index:
            route = await self.get_route(address)
            if route is None:
                return None
            index = route.get_attr(route.name2nla("oif"))
        try:
            return self.ip.neigh("get", dst=str(address), ifindex=index, state="none")[0]
        except pyroute2.netlink.exceptions.NetlinkError as e:
            if e.code == errno.ENOENT:
                return None
            raise

    async def get_neighbor_mac(
        self,
        address: cptypes.IPAddress,
        *,
        index: int = 0,  # interface index
        flags: int = 0,
    ) -> cptypes.MacAddress | None:
        neigh = await self.get_neighbor(address, index=index, flags=flags)
        if neigh is None:
            return None
        mac = neigh.get_attr(neigh.name2nla("lladdr"))
        if mac is None:
            return None
        return cptypes.MacAddress.parse(mac)

    async def get_route(
        self,
        address: cptypes.IPAddress,
    ) -> pyroute2.iproute.linux.rtmsg | None:
        try:
            return self.ip.route("get", dst=str(address))[0]
        except pyroute2.netlink.exceptions.NetlinkError as e:
            if e.code == errno.ENOENT:
                return None
            raise

    async def dump_neighbors(
        self,
        interface: str,
    ) -> typing.AsyncGenerator[tuple[cptypes.MacAddress, cptypes.IPAddress], None]:
        ifindex = socket.if_nametoindex(interface)
        unicast_num = pyroute2.netlink.rtnl.rt_type["unicast"]
        # ip.neigh doesn't support AF_UNSPEC (as it is 0 and evaluates to `False` and gets forced to AF_INET)
        for family in (socket.AF_INET, socket.AF_INET6):
            for neigh in self.ip.neigh("dump", ifindex=ifindex, family=family):
                if neigh["ndm_type"] != unicast_num:
                    continue
                mac = neigh.get_attr(neigh.name2nla("lladdr"))
                if not mac:
                    continue
                dst = ipaddress.ip_address(neigh.get_attr(neigh.name2nla("dst")))
                if dst.is_link_local:
                    continue
                yield (cptypes.MacAddress.parse(mac), dst)
