from __future__ import annotations

import typing

import pyroute2.netlink  # type: ignore
from pyroute2.netlink.nfnetlink import nftsocket as _nftsocket  # type: ignore

from capport import cptypes

from .nft_socket import NFTSocket

NFPROTO_INET: int = 1  # nfgen_family "ipv4+ipv6"; strace decodes this as "AF_UNIX"


def _from_msec(msecs: int | None) -> float | None:
    # to seconds
    if msecs is None:
        return None
    return msecs / 1000.0


class NftSet:
    def __init__(self):
        self._socket = NFTSocket()
        self._socket.bind()

    @staticmethod
    def _set_elem(
        mac: cptypes.MacAddress,
        timeout: int | float | None = None,
    ) -> _nftsocket.nft_set_elem_list_msg.set_elem:
        attrs: dict[str, typing.Any] = {
            "NFTA_SET_ELEM_KEY": dict(
                NFTA_DATA_VALUE=mac.raw,
            ),
        }
        if timeout:
            attrs["NFTA_SET_ELEM_TIMEOUT"] = int(1000 * timeout)
        return attrs

    def _bulk_insert(
        self,
        entries: typing.Sequence[tuple[cptypes.MacAddress, int | float]],
    ) -> None:
        ser_entries = [self._set_elem(mac) for mac, _timeout in entries]
        ser_entries_with_timeout = [self._set_elem(mac, timeout) for mac, timeout in entries]
        with self._socket.begin() as tx:
            # create doesn't affect existing elements, so:
            # make sure entries exists
            tx.put(
                _nftsocket.NFT_MSG_NEWSETELEM,
                pyroute2.netlink.NLM_F_CREATE,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                    NFTA_SET_ELEM_LIST_SET="allowed",
                    NFTA_SET_ELEM_LIST_ELEMENTS=ser_entries,
                ),
            )
            # drop entries (would fail if it doesn't exist)
            tx.put(
                _nftsocket.NFT_MSG_DELSETELEM,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                    NFTA_SET_ELEM_LIST_SET="allowed",
                    NFTA_SET_ELEM_LIST_ELEMENTS=ser_entries,
                ),
            )
            # now create entries with new timeout value
            tx.put(
                _nftsocket.NFT_MSG_NEWSETELEM,
                pyroute2.netlink.NLM_F_CREATE | pyroute2.netlink.NLM_F_EXCL,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                    NFTA_SET_ELEM_LIST_SET="allowed",
                    NFTA_SET_ELEM_LIST_ELEMENTS=ser_entries_with_timeout,
                ),
            )

    def bulk_insert(self, entries: typing.Sequence[tuple[cptypes.MacAddress, int | float]]) -> None:
        # limit chunk size
        while len(entries) > 0:
            self._bulk_insert(entries[:1024])
            entries = entries[1024:]

    def insert(self, mac: cptypes.MacAddress, timeout: int | float) -> None:
        self.bulk_insert([(mac, timeout)])

    def _bulk_remove(self, entries: typing.Sequence[cptypes.MacAddress]) -> None:
        ser_entries = [self._set_elem(mac) for mac in entries]
        with self._socket.begin() as tx:
            # make sure entries exists
            tx.put(
                _nftsocket.NFT_MSG_NEWSETELEM,
                pyroute2.netlink.NLM_F_CREATE,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                    NFTA_SET_ELEM_LIST_SET="allowed",
                    NFTA_SET_ELEM_LIST_ELEMENTS=ser_entries,
                ),
            )
            # drop entries (would fail if it doesn't exist)
            tx.put(
                _nftsocket.NFT_MSG_DELSETELEM,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                    NFTA_SET_ELEM_LIST_SET="allowed",
                    NFTA_SET_ELEM_LIST_ELEMENTS=ser_entries,
                ),
            )

    def bulk_remove(self, entries: typing.Sequence[cptypes.MacAddress]) -> None:
        # limit chunk size
        while len(entries) > 0:
            self._bulk_remove(entries[:1024])
            entries = entries[1024:]

    def remove(self, mac: cptypes.MacAddress) -> None:
        self.bulk_remove([mac])

    def list(self) -> list:
        responses: typing.Iterator[_nftsocket.nft_set_elem_list_msg]
        responses = self._socket.nft_dump(
            _nftsocket.NFT_MSG_GETSETELEM,
            nfgen_family=NFPROTO_INET,
            attrs=dict(
                NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                NFTA_SET_ELEM_LIST_SET="allowed",
            ),
        )
        return [
            {
                "mac": cptypes.MacAddress(
                    elem.get_attr("NFTA_SET_ELEM_KEY").get_attr("NFTA_DATA_VALUE"),
                ),
                "timeout": _from_msec(elem.get_attr("NFTA_SET_ELEM_TIMEOUT", None)),
                "expiration": _from_msec(elem.get_attr("NFTA_SET_ELEM_EXPIRATION", None)),
            }
            for response in responses
            for elem in response.get_attr("NFTA_SET_ELEM_LIST_ELEMENTS", [])
        ]

    def flush(self) -> None:
        self._socket.nft_put(
            _nftsocket.NFT_MSG_DELSETELEM,
            nfgen_family=NFPROTO_INET,
            attrs=dict(
                NFTA_SET_ELEM_LIST_TABLE="captive_mark",
                NFTA_SET_ELEM_LIST_SET="allowed",
            ),
        )

    def create(self):
        with self._socket.begin() as tx:
            tx.put(
                _nftsocket.NFT_MSG_NEWTABLE,
                pyroute2.netlink.NLM_F_CREATE,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_TABLE_NAME="captive_mark",
                ),
            )
            tx.put(
                _nftsocket.NFT_MSG_NEWSET,
                pyroute2.netlink.NLM_F_CREATE,
                nfgen_family=NFPROTO_INET,
                attrs=dict(
                    NFTA_SET_TABLE="captive_mark",
                    NFTA_SET_NAME="allowed",
                    NFTA_SET_FLAGS=0x10,  # NFT_SET_TIMEOUT
                    NFTA_SET_KEY_TYPE=9,  # nft type for "type ether_addr" - only relevant for userspace nft
                    NFTA_SET_KEY_LEN=6,  # length of key: mac address
                    NFTA_SET_ID=1,  # kernel seems to need a set id unique per transaction
                ),
            )
