from __future__ import annotations

import contextlib
import os
import socket
import typing

import trio
import trio.socket


def _check_watchdog_pid() -> bool:
    wpid = os.environ.pop("WATCHDOG_PID", None)
    if not wpid:
        return True
    return wpid == str(os.getpid())


@contextlib.asynccontextmanager
async def open_sdnotify() -> typing.AsyncGenerator[SdNotify, None]:
    target = os.environ.pop("NOTIFY_SOCKET", None)
    ns: trio.socket.SocketType | None = None
    watchdog_usec: int = 0
    if target:
        if target.startswith("@"):
            # Linux abstract namespace socket
            target = "\0" + target[1:]
        ns = trio.socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        await ns.connect(target)
        watchdog_usec_s = os.environ.pop("WATCHDOG_USEC", None)
        if _check_watchdog_pid() and watchdog_usec_s:
            watchdog_usec = int(watchdog_usec_s)
    try:
        async with trio.open_nursery() as nursery:
            sn = SdNotify(_ns=ns)
            if watchdog_usec:
                await nursery.start(sn._run_watchdog, watchdog_usec)
            yield sn
            # stop watchdoch
            nursery.cancel_scope.cancel()
    finally:
        if ns:
            ns.close()


class SdNotify:
    def __init__(self, *, _ns: trio.socket.SocketType | None) -> None:
        self._ns = _ns

    def is_connected(self) -> bool:
        return not (self._ns is None)

    async def _run_watchdog(self, watchdog_usec: int, *, task_status=trio.TASK_STATUS_IGNORED) -> None:
        assert self.is_connected(), "Watchdog can't run without socket"
        await self.send("WATCHDOG=1")
        task_status.started()
        # send every half of the watchdog timeout
        interval = (watchdog_usec / 1e6) / 2.0
        while True:
            await trio.sleep(interval)
            await self.send("WATCHDOG=1")

    async def send(self, *msg: str) -> None:
        if not self.is_connected():
            return
        dgram = "\n".join(msg).encode("utf-8")
        assert self._ns, "not connected"  # checked above
        sent = await self._ns.send(dgram)
        if sent != len(dgram):
            raise OSError("Sent incomplete datagram to NOTIFY_SOCKET")
